import docker, dockerpty, os
from ipaddress import IPv4Network, IPv4Address
import socket
from . import api


_conns = []

def next_ip(network_name, context):
    client = docker.Client(version="auto")
    network = client.inspect_network(network_name)

    subnet = IPv4Network(network['IPAM']['Config'][0]['Subnet'])
    reserved = [IPv4Address(v['IPv4Address'].split('/')[0]) for k,v in network['Containers'].items()]
    
    subnet_hosts = subnet.hosts()
    next(subnet_hosts) #skip first address
    for address in subnet_hosts:
        if address in reserved: continue
        
        return str(address)

    raise Exception("Could not find unused address in subnet {}".format(subnet))


def random_open_port(context):
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    _conns.append(tcp)
    tcp.bind(('', 0))
    addr, port = tcp.getsockname()
    
    return str(port)


def load_properties(paths, context):
    if type(paths) is not list:
        paths = [paths,]
    
    file_found = False
    for path in paths:
        try:
            if context:
                properties = api.load_properties(os.path.join(os.path.dirname(context.file_location), path))
            else:
                properties = api.load_properties(path)

            file_found = True
        except FileNotFoundError:
            pass

    if not file_found:
        raise FileNotFoundError("None of the following properties files were found: {}".format(','.join(paths)))
    
    return properties


def before_container_start():
    for conn in _conns:
        conn.close()    


def global_wrap(func, context):
    def wrapper(*args, **kwargs):
        return func(*args, **{**kwargs, **{ 'context': context } })

    return wrapper


def wrap_globals(context):
    wrapped_globals = {}
    for key in globals.keys():
        wrapped_globals[key] = global_wrap(globals[key], context)

    return wrapped_globals


globals = {
    'next_ip': next_ip,
    'random_open_port': random_open_port,
    'load_properties': load_properties
}


from . import configuration, process_manager, api
import tempfile, jinja2
from distutils import dir_util
import os, yaml, ast, sys, jinja2, time
from .properties import Properties
from .context_globals import wrap_globals
import templated_yaml.api as tapi


def run(config_path, command='dev', logger=lambda x: None):
    abs_config_path = os.path.abspath(config_path)
    config = configuration.Config.load_from(abs_config_path)
    tmp_dir = tempfile.mkdtemp(dir='/tmp')
    dir_util.copy_tree(
        os.path.join(os.path.dirname(__file__), 'docker'),
        tmp_dir
    )

    config.volumes.append('{}:/docker'.format(tmp_dir))

    process_manager.build_image(config, logger)
    return process_manager.run_image(config, command, logger)


def create_configuration(config_path, template_path):
    abs_config_path = os.path.abspath(config_path)
    abs_template_path = os.path.abspath(template_path)

    context = tapi.render_from_path(abs_config_path, globals=wrap_globals(None))

    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(os.path.dirname(abs_template_path)))
    template = jinja_env.get_template(os.path.basename(abs_template_path))

    return template.render(context)


def load_properties(file_location):
    p = Properties()
    p.load(open(file_location))

    return p.getPropertyDict()


def create_init_file(config_path, format):
    abs_config_path = os.path.abspath(config_path)
    config = configuration.Config.load_from(abs_config_path)

    here = os.path.dirname(os.path.abspath(__file__))

    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(os.path.join(here, 'templates')))
    template = jinja_env.get_template('{}.j2'.format(format))
    
    return template.render({
        'wharf_file_location': abs_config_path,
        'image_name': config.name
    })


import templated_yaml.api as tapi
from collections import namedtuple
from wharf import settings
from .context_globals import wrap_globals


ImageConfig = namedtuple('ImageConfig', ['repository', 'tag'])
NetworkConfig = namedtuple('NetworkConfig', ['name', 'ipv4_address', 'aliases', 'links'])
PortMapping = namedtuple('PortMapping', ['host', 'container'])
WharfConfig = namedtuple('WharfConfig', ['min_version'])
BuildContextConfig = namedtuple('BuildContextConfig', ['excludes'])

class ConfigurationError(Exception):
    pass

class Config(object):

    @classmethod
    def load_from(cls, path):
        config = Config()
        config._file_location = path
        config._data = tapi.render_from_path(path, globals=wrap_globals(config))
        config._data.setdefault('volumes', [])
        config._data.setdefault('dockerignore', [])
        config._data.setdefault('environment', {})
        config._data.setdefault('port_mappings', [])

        if config.wharf.min_version > (0, 0, 0):
            if config.wharf.min_version > settings.VERSION:
                raise ConfigurationError("This configuration requires at least wharf version " + str(config.wharf.min_version))

        return config

    @property
    def file_location(self):
        return self._file_location

    @property
    def dockerfile(self):
        return self._data.get('dockerfile')

    @property
    def image(self):
        image_data = self._data.get('image')
        return ImageConfig(image_data.get('repository'), image_data.get('tag'))

    @property
    def environment(self):
        return self._data.get('environment', {})

    @property
    def volumes(self):
        return self._data.get('volumes')

    @property
    def port_mappings(self):
        mappings = self._data.get('port_mappings', [])
        parsed_mappings = []
        for mapping in mappings:
            parsed_mappings.append(
                PortMapping(mapping.get('host'), mapping.get('container'))
            )

        return parsed_mappings

    @property
    def entrypoint(self):
        return self._data.get('entrypoint', None)

    @property
    def hostname(self):
        return self._data.get('hostname', None)

    @property
    def name(self):
        return self._data.get('name', None)

    @property
    def extra_hosts(self):
        return self._data.get('extra_hosts', None)

    @property
    def wharf(self):
        wharf_node = self._data.get('wharf', {})
        min_version = wharf_node.get('min_version', (0,0,0))
        if not isinstance(min_version, tuple):
            raise ConfigurationError("wharf.min_version must be a tuple in the format (major,minor,patch)")

        config = (
            ( 'min_version', min_version ),
        )

        return WharfConfig( *[c[1] for c in config] )

    @property
    def dockerignore(self):
        return self._data.get('dockerignore', [])

    @property
    def network(self):
        network = self._data.get('network', None)
        network_config = None

        if network:
            network_config = NetworkConfig(
                network['name'],
                network['ipv4_address'],
                network.get('aliases', []),
                network.get('links', [])
            )

        return network_config





import docker, dockerpty, os
from io import BytesIO
import json
import wharf.context_globals


def build_image(config, logger=lambda x: None):
    client = docker.Client(version="auto")
    encoded_dockerfile = BytesIO(config.dockerfile.encode('utf-8'))

    context = docker.utils.tar(os.path.dirname(config.file_location), exclude=config.dockerignore)
    context.seek(0)

    import tarfile
    context_tar = tarfile.open(mode='a', fileobj=context)
    dfinfo = tarfile.TarInfo('Dockerfile')
    dfinfo.size = len(encoded_dockerfile.getvalue())
    context_tar.addfile(dfinfo, encoded_dockerfile)
    context_tar.close()
    context.seek(0)

    for output in client.build(
        fileobj=context,
        custom_context=True,
        rm=True,
        tag="{}:{}".format(config.image.repository, config.image.tag)
    ):
        # Bug with how the output is returned:  several dicts may be returned at once and they
        # can't be decoded until you split them.
        lines = [json.loads(l) for l in output.decode('utf-8').split('\r\n') if len(l.strip())]
        error = False
        for line in lines:
            if line.get('stream'):
                logger(line['stream'].replace('\n', ''))
            elif line.get('status'):
                logger(line['status'].replace('\n', ''))
            else:
                error = True
                logger(line.get('errorDetail', {}).get('message').replace('\n', ''))

    if error:
        raise Exception("Failed to build image")


def run_image(config, command, logger=lambda x: None):
    client = docker.Client(version="auto")

    all_volumes = config.volumes
    volumes = [v.split(':')[1] for v in all_volumes]

    binds = []
    for volume in all_volumes:
        parts = volume.split(':')
        host_path = parts[0]

        if not os.path.isabs(host_path):
            host_path = os.path.realpath(
                os.path.join(
                    os.path.dirname(config.file_location),
                    os.path.expanduser(host_path)
                )
            )

        parts[0] = host_path
        binds.append(":".join(parts))

    env = {
        'LOCAL_USER_ID': os.getuid()
    }
    env.update(config.environment)

    port_bindings = { }
    for mapping in config.port_mappings:
        port_bindings[mapping.container] = mapping.host

    if config.name:
        try:
            client.remove_container(config.name, force=True)
        except docker.errors.NotFound:
            pass

    networking_config = None
    if config.network:
        networking_config = client.create_networking_config({
            config.network.name: client.create_endpoint_config(
                ipv4_address=config.network.ipv4_address,
                aliases=config.network.aliases,
                links=config.network.links
            )
        })


    container = client.create_container(
        image='{}:{}'.format(config.image.repository, config.image.tag),
        stdin_open=True,
        tty=True,
        command=command,
        volumes=volumes,
        host_config=client.create_host_config(
            binds=binds,
            port_bindings=port_bindings,
            extra_hosts=config.extra_hosts
        ),
        environment=env,
        entrypoint=config.entrypoint,
        name=config.name,
        hostname=config.hostname,
        networking_config=networking_config
    )

    wharf.context_globals.before_container_start()
    pty = dockerpty.start(client, container)
    exit_code = int(client.inspect_container(container).get('State', {}).get('ExitCode', 0)) 
    
    return exit_code


import click
from . import settings, api


@click.group()
@click.pass_context
def cli(ctx):
    pass


@cli.command()
@click.argument('config_path')
@click.argument('command', default='dev')
def run(config_path, command='dev'):
    exit_code = api.run(config_path, command, click.echo)
    exit(exit_code)


@cli.command()
@click.argument("config_path")
@click.argument("template")
def create_configuration(config_path, template):
    config = api.create_configuration(config_path, template)
    click.echo(config)


@cli.command()
@click.argument("config_path")
@click.option('--format', type=click.Choice(['systemd']), default='systemd')
def create_init_file(config_path, format):
    config = api.create_init_file(config_path, format)
    click.echo(config)


@cli.command()
def version():
    major, minor, patch = settings.VERSION

    click.echo("Wharf version {}.{}.{} compiled on {}".format(
        major,
        minor,
        patch,
        settings.COMPILATION_DATE.strftime('%Y-%m-%d %H:%M:%S')
    ))


